@@ -6,10 +6,11 @@ DROP TABLE build_notifications;
 DROP TABLE log_messages;
 
 DROP TABLE buildroot_listing;
-DROP TABLE imageinfo_listing;
+DROP TABLE image_listing;
 
 DROP TABLE rpminfo;
-DROP TABLE imageinfo;
+DROP TABLE image_builds;
+DROP TABLE image_archives;
 
 DROP TABLE group_package_listing;
 DROP TABLE group_req_listing;
@@ -101,6 +102,8 @@ INSERT INTO permissions (name) VALUES ('build');
 INSERT INTO permissions (name) VALUES ('repo');
 INSERT INTO permissions (name) VALUES ('livecd');
 INSERT INTO permissions (name) VALUES ('maven-import');
+INSERT INTO permissions (name) VALUES ('win-import');
+INSERT INTO permissions (name) VALUES ('win-admin');
 INSERT INTO permissions (name) VALUES ('appliance');
 
 CREATE TABLE user_perms (
@@ -179,6 +182,7 @@ INSERT INTO channels (name) VALUES ('createrepo');
 INSERT INTO channels (name) VALUES ('maven');
 INSERT INTO channels (name) VALUES ('livecd');
 INSERT INTO channels (name) VALUES ('appliance');
+INSERT INTO channels (name) VALUES ('vm');
 
 -- Here we track the build machines
 -- each host has an entry in the users table also
@@ -258,6 +262,13 @@ CREATE TABLE package (
 -- (implicitly created by unique constraint)
 
 
+CREATE TABLE volume (
+        id SERIAL NOT NULL PRIMARY KEY,
+        name TEXT UNIQUE NOT NULL
+) WITHOUT OIDS;
+
+INSERT INTO volume (id, name) VALUES (0, 'DEFAULT');
+
 -- here we track the built packages
 -- this is at the srpm level, since builds are by srpm
 -- see rpminfo for isolated packages
@@ -266,6 +277,7 @@ CREATE TABLE package (
 --   null, or may point to a deleted task.
 CREATE TABLE build (
 	id SERIAL NOT NULL PRIMARY KEY,
+        volume_id INTEGER NOT NULL REFERENCES volume (id),
 	pkg_id INTEGER NOT NULL REFERENCES package (id) DEFERRABLE,
 	version TEXT NOT NULL,
 	release TEXT NOT NULL,
@@ -359,6 +371,20 @@ CREATE TABLE tag_config (
 ) WITHOUT OIDS;
 
 
+-- the tag_updates table provides a mechanism to indicate changes relevant to tag
+-- that are not reflected in a versioned table. For example: builds changing volumes,
+-- changes to external repo content, additional rpms imported to an existing build
+CREATE TABLE tag_updates (
+        id SERIAL NOT NULL PRIMARY KEY,
+        tag_id INTEGER NOT NULL REFERENCES tag(id),
+        update_event INTEGER NOT NULL REFERENCES events(id) DEFAULT get_event(),
+        updater_id INTEGER NOT NULL REFERENCES users(id),
+        update_type INTEGER NOT NULL
+) WITHOUT OIDS;
+
+CREATE INDEX tag_updates_by_tag ON tag_updates (tag_id);
+CREATE INDEX tag_updates_by_event ON tag_updates (update_event);
+
 -- a build target tells the system where to build the package
 -- and how to tag it afterwards.
 CREATE TABLE build_target (
@@ -449,16 +475,9 @@ CREATE TABLE buildroot (
 ) WITHOUT OIDS;
 
 -- track spun images (livecds, installation, VMs...)
-CREATE TABLE imageinfo (
-	id SERIAL NOT NULL PRIMARY KEY,
-        task_id INTEGER NOT NULL REFERENCES task(id),
-	filename TEXT NOT NULL,
-	filesize BIGINT NOT NULL,
-	arch VARCHAR(16) NOT NULL,
-	hash TEXT NOT NULL,
-	mediatype TEXT NOT NULL
+CREATE TABLE image_builds (
+    build_id INTEGER NOT NULL PRIMARY KEY REFERENCES build(id)
 ) WITHOUT OIDS;
-CREATE INDEX imageinfo_task_id on imageinfo(task_id);
 
 -- this table associates tags with builds.  an entry here tags a package
 CREATE TABLE tag_listing (
@@ -620,14 +639,6 @@ CREATE TABLE buildroot_listing (
 ) WITHOUT OIDS;
 CREATE INDEX buildroot_listing_rpms ON buildroot_listing(rpm_id);
 
--- tracks the contents of an image
-CREATE TABLE imageinfo_listing (
-	image_id INTEGER NOT NULL REFERENCES imageinfo(id),
-	rpm_id INTEGER NOT NULL REFERENCES rpminfo(id),
-	UNIQUE (image_id, rpm_id)
-) WITHOUT OIDS;
-CREATE INDEX imageinfo_listing_rpms on imageinfo_listing(rpm_id);
-
 CREATE TABLE log_messages (
     id SERIAL NOT NULL PRIMARY KEY,
     message TEXT NOT NULL,
@@ -668,6 +679,12 @@ CREATE TABLE maven_builds (
         version TEXT NOT NULL
 ) WITHOUT OIDS;
 
+-- Windows-specific build information
+CREATE TABLE win_builds (
+        build_id INTEGER NOT NULL PRIMARY KEY REFERENCES build(id),
+        platform TEXT NOT NULL
+) WITHOUT OIDS;
+
 -- Even though we call this archiveinfo, we can probably use it for
 -- any filetype output by a build process.  In general they will be
 -- archives (.zip, .jar, .tar.gz) but could also be installer executables (.exe)
@@ -678,11 +695,30 @@ CREATE TABLE archivetypes (
         extensions TEXT NOT NULL
 ) WITHOUT OIDS;
 
-insert into archivetypes (name, description, extensions) values ('jar', 'Jar files', 'jar war rar ear');
-insert into archivetypes (name, description, extensions) values ('zip', 'Zip archives', 'zip');
-insert into archivetypes (name, description, extensions) values ('pom', 'Maven Project Object Management files', 'pom');
-insert into archivetypes (name, description, extensions) values ('tar', 'Tar files', 'tar tar.gz tar.bz2');
-insert into archivetypes (name, description, extensions) values ('xml', 'XML files', 'xml');
+insert into archivetypes (name, description, extensions) values ('jar', 'Jar file', 'jar war rar ear');
+insert into archivetypes (name, description, extensions) values ('zip', 'Zip archive', 'zip');
+insert into archivetypes (name, description, extensions) values ('pom', 'Maven Project Object Management file', 'pom');
+insert into archivetypes (name, description, extensions) values ('tar', 'Tar file', 'tar tar.gz tar.bz2');
+insert into archivetypes (name, description, extensions) values ('xml', 'XML file', 'xml');
+insert into archivetypes (name, description, extensions) values ('xsd', 'XML Schema Definition', 'xsd');
+insert into archivetypes (name, description, extensions) values ('spec', 'RPM spec file', 'spec');
+insert into archivetypes (name, description, extensions) values ('exe', 'Windows executable', 'exe');
+insert into archivetypes (name, description, extensions) values ('dll', 'Windows dynamic link library', 'dll');
+insert into archivetypes (name, description, extensions) values ('lib', 'Windows import library', 'lib');
+insert into archivetypes (name, description, extensions) values ('sys', 'Windows device driver', 'sys');
+insert into archivetypes (name, description, extensions) values ('inf', 'Windows driver information file', 'inf');
+insert into archivetypes (name, description, extensions) values ('cat', 'Windows catalog file', 'cat');
+insert into archivetypes (name, description, extensions) values ('msi', 'Windows Installer package', 'msi');
+insert into archivetypes (name, description, extensions) values ('pdb', 'Windows debug information', 'pdb');
+insert into archivetypes (name, description, extensions) values ('oem', 'Windows driver oem file', 'oem');
+insert into archivetypes (name, description, extensions) values ('iso', 'CD/DVD Image', 'iso');
+insert into archivetypes (name, description, extensions) values ('raw', 'Raw disk image', 'raw');
+insert into archivetypes (name, description, extensions) values ('qcow', 'QCOW image', 'qcow');
+insert into archivetypes (name, description, extensions) values ('qcow2', 'QCOW2 image', 'qcow2');
+insert into archivetypes (name, description, extensions) values ('vmdk', 'vSphere image', 'vmdk');
+insert into archivetypes (name, description, extensions) values ('ova', 'OVA image', 'ova');
+insert into archivetypes (name, description, extensions) values ('ks', 'Kickstart', 'ks');
+insert into archivetypes (name, description, extensions) values ('cfg', 'Configuration file', 'cfg');
 
 -- Do we want to enforce a constraint that a build can only generate one
 -- archive with a given name?
@@ -692,8 +728,9 @@ CREATE TABLE archiveinfo (
 	build_id INTEGER NOT NULL REFERENCES build (id),
 	buildroot_id INTEGER REFERENCES buildroot (id),
 	filename TEXT NOT NULL,
-	size INTEGER NOT NULL,
-	md5sum TEXT NOT NULL
+	size BIGINT NOT NULL,
+	checksum TEXT NOT NULL,
+	checksum_type INTEGER NOT NULL
 ) WITHOUT OIDS;
 CREATE INDEX archiveinfo_build_idx ON archiveinfo (build_id);
 CREATE INDEX archiveinfo_buildroot_idx on archiveinfo (buildroot_id);
@@ -707,6 +744,19 @@ CREATE TABLE maven_archives (
         version TEXT NOT NULL
 ) WITHOUT OIDS;
 
+CREATE TABLE image_archives (
+    archive_id INTEGER NOT NULL PRIMARY KEY REFERENCES archiveinfo(id),
+    arch VARCHAR(16) NOT NULL
+) WITHOUT OIDS;
+
+-- tracks the contents of an image
+CREATE TABLE image_listing (
+	image_id INTEGER NOT NULL REFERENCES image_archives(archive_id),
+	rpm_id INTEGER NOT NULL REFERENCES rpminfo(id),
+	UNIQUE (image_id, rpm_id)
+) WITHOUT OIDS;
+CREATE INDEX image_listing_rpms on image_listing(rpm_id);
+
 CREATE TABLE buildroot_archives (
 	buildroot_id INTEGER NOT NULL REFERENCES buildroot (id),
 	archive_id INTEGER NOT NULL REFERENCES archiveinfo (id),
@@ -715,4 +765,12 @@ CREATE TABLE buildroot_archives (
 ) WITHOUT OIDS;
 CREATE INDEX buildroot_archives_archive_idx ON buildroot_archives (archive_id);
 
+-- Extended information about files built in Windows VMs
+CREATE TABLE win_archives (
+        archive_id INTEGER NOT NULL PRIMARY KEY REFERENCES archiveinfo(id),
+        relpath TEXT NOT NULL,
+        platforms TEXT NOT NULL,
+        flags TEXT
+) WITHOUT OIDS;
+
 COMMIT WORK;
