@@ -1,7 +1,5 @@
-# mod_python script
-
 # kojixmlrpc - an XMLRPC interface for koji.
-# Copyright (c) 2005-2007 Red Hat
+# Copyright (c) 2005-2012 Red Hat
 #
 #    Koji is free software; you can redistribute it and/or
 #    modify it under the terms of the GNU Lesser General Public
@@ -20,24 +18,25 @@
 # Authors:
 #       Mike McLean <mikem@redhat.com>
 
-from ConfigParser import ConfigParser
+from ConfigParser import RawConfigParser
+import inspect
 import logging
+import os
 import sys
 import time
 import traceback
 import types
 import pprint
-from xmlrpclib import loads,dumps,Fault
-from mod_python import apache
+import resource
+from xmlrpclib import getparser,dumps,Fault
+from koji.server import WSGIWrapper
 
 import koji
 import koji.auth
 import koji.db
 import koji.plugin
 import koji.policy
-import kojihub
-from kojihub import RootExports
-from kojihub import HostExports
+import koji.util
 from koji.context import context
 
 
@@ -51,6 +50,7 @@ class HandlerRegistry(object):
         self.register_function(self.system_listMethods, name="system.listMethods")
         self.register_function(self.system_methodSignature, name="system.methodSignature")
         self.register_function(self.system_methodHelp, name="system.methodHelp")
+        self.argspec_cache = {}
 
     def register_function(self, function, name = None):
         if name is None:
@@ -100,14 +100,33 @@ class HandlerRegistry(object):
                     for cbtype in v.callbacks:
                         koji.plugin.register_callback(cbtype, v)
 
+    def getargspec(self, func):
+        ret = self.argspec_cache.get(func)
+        if ret:
+            return ret
+        ret = tuple(inspect.getargspec(func))
+        if inspect.ismethod(func) and func.im_self:
+            # bound method, remove first arg
+            args, varargs, varkw, defaults = ret
+            if args:
+                aname = args[0] #generally "self"
+                del args[0]
+                if defaults and aname in defaults:
+                    # shouldn't happen, but...
+                    del defaults[aname]
+        return ret
+
     def list_api(self):
         funcs = []
         for name,func in self.funcs.items():
             #the keys in self.funcs determine the name of the method as seen over xmlrpc
             #func.__name__ might differ (e.g. for dotted method names)
             args = self._getFuncArgs(func)
+            argspec = self.getargspec(func)
             funcs.append({'name': name,
                           'doc': func.__doc__,
+                          'argspec': argspec,
+                          'argdesc': inspect.formatargspec(*argspec),
                           'args': args})
         return funcs
 
@@ -133,13 +152,8 @@ class HandlerRegistry(object):
         func = self.funcs.get(method)
         if func is None:
             return ""
-        arglist = []
-        for arg in self._getFuncArgs(func):
-            if isinstance(arg,str):
-                arglist.append(arg)
-            else:
-                arglist.append('%s=%s' % (arg[0], arg[1]))
-        ret = '%s(%s)' % (method, ", ".join(arglist))
+        args = inspect.formatargspec(*self.getargspec(func))
+        ret = '%s%s' % (method, args)
         if func.__doc__:
             ret += "\ndescription: %s" % func.__doc__
         return ret
@@ -180,15 +194,22 @@ class ModXMLRPCRequestHandler(object):
         else:
             return self.handlers.get(name)
 
-    def _marshaled_dispatch(self, data):
-        """Dispatches an XML-RPC method from marshalled (XML) data."""
+    def _read_request(self, stream):
+        parser, unmarshaller = getparser()
+        while True:
+            chunk = stream.read(8192)
+            if not chunk:
+                break
+            parser.feed(chunk)
+        parser.close()
+        return unmarshaller.close(), unmarshaller.getmethodname()
 
-        params, method = loads(data)
+    def _wrap_handler(self, handler, environ):
+        """Catch exceptions and encode response of handler"""
 
-        start = time.time()
         # generate response
         try:
-            response = self._dispatch(method, params)
+            response = handler(environ)
             # wrap response in a singleton tuple
             response = (response,)
             response = dumps(response, methodresponse=1, allow_none=1)
@@ -222,9 +243,18 @@ class ModXMLRPCRequestHandler(object):
 
         return response
 
-    def _dispatch(self,method,params):
-        func = self._get_handler(method)
-        context.method = method
+    def handle_upload(self, environ):
+        #uploads can't be in a multicall
+        context.method = None
+        self.check_session()
+        self.enforce_lockout()
+        return kojihub.handle_upload(environ)
+
+    def handle_rpc(self, environ):
+        params, method = self._read_request(environ['wsgi.input'])
+        return self._dispatch(method, params)
+
+    def check_session(self):
         if not hasattr(context,"session"):
             #we may be called again by one of our meta-calls (like multiCall)
             #so we should only create a session if one does not already exist
@@ -233,29 +263,40 @@ class ModXMLRPCRequestHandler(object):
                 context.session.validate()
             except koji.AuthLockError:
                 #might be ok, depending on method
-                if method not in ('exclusiveSession','login', 'krbLogin', 'logout'):
+                if context.method not in ('exclusiveSession','login', 'krbLogin', 'logout'):
                     raise
-            if context.opts.get('LockOut') and \
-                   method not in ('login', 'krbLogin', 'sslLogin', 'logout'):
-                if not context.session.hasPerm('admin'):
-                    raise koji.ServerOffline, "Server disabled for maintenance"
-        # handle named parameters
-        params,opts = koji.decode_args(*params)
 
-        if self.logger.isEnabledFor(logging.DEBUG):
-            self.logger.debug("Handling method %s for session %s (#%s)",
+    def enforce_lockout(self):
+        if context.opts.get('LockOut') and \
+            context.method not in ('login', 'krbLogin', 'sslLogin', 'logout') and \
+            not context.session.hasPerm('admin'):
+                raise koji.ServerOffline, "Server disabled for maintenance"
+
+    def _dispatch(self, method, params):
+        func = self._get_handler(method)
+        context.method = method
+        context.params = params
+        self.check_session()
+        self.enforce_lockout()
+        # handle named parameters
+        params, opts = koji.decode_args(*params)
+
+        if self.logger.isEnabledFor(logging.INFO):
+            self.logger.info("Handling method %s for session %s (#%s)",
                             method, context.session.id, context.session.callnum)
-            if method != 'uploadFile':
+            if method != 'uploadFile' and self.logger.isEnabledFor(logging.DEBUG):
                 self.logger.debug("Params: %s", pprint.pformat(params))
                 self.logger.debug("Opts: %s", pprint.pformat(opts))
             start = time.time()
 
-        ret = func(*params,**opts)
+        ret = koji.util.call_with_argcheck(func, params, opts)
 
-        if self.logger.isEnabledFor(logging.DEBUG):
-            self.logger.debug("Completed method %s for session %s (#%s): %f seconds",
+        if self.logger.isEnabledFor(logging.INFO):
+            rusage = resource.getrusage(resource.RUSAGE_SELF)
+            self.logger.info("Completed method %s for session %s (#%s): %f seconds, rss %s, stime %f",
                             method, context.session.id, context.session.callnum,
-                            time.time()-start)
+                            time.time()-start,
+                            rusage.ru_maxrss, rusage.ru_stime)
 
         return ret
 
@@ -286,40 +327,11 @@ class ModXMLRPCRequestHandler(object):
     def handle_request(self,req):
         """Handle a single XML-RPC request"""
 
-        start = time.time()
-        # XMLRPC uses POST only. Reject anything else
-        if req.method != 'POST':
-            req.allow_methods(['POST'],1)
-            raise apache.SERVER_RETURN, apache.HTTP_METHOD_NOT_ALLOWED
-
-        response = self._marshaled_dispatch(req.read())
-
-        req.content_type = "text/xml"
-        req.set_content_length(len(response))
-        req.write(response)
-        self.logger.debug("Returning %d bytes after %f seconds", len(response),
-                        time.time() - start)
+        pass
+        #XXX no longer used
 
 
-
-def dump_req(req):
-    data = [
-        "request: %s\n" % req.the_request,
-        "uri: %s\n" % req.uri,
-        "args: %s\n" % req.args,
-        "protocol: %s\n" % req.protocol,
-        "method: %s\n" % req.method,
-        "https: %s\n" % req.is_https(),
-        "auth_name: %s\n" % req.auth_name(),
-        "auth_type: %s\n" % req.auth_type(),
-        "headers:\n%s\n" % pprint.pformat(req.headers_in),
-    ]
-    for part in data:
-        sys.stderr.write(part)
-    sys.stderr.flush()
-
-
-def offline_reply(req, msg=None):
+def offline_reply(start_response, msg=None):
     """Send a ServerOffline reply"""
     faultCode = koji.ServerOffline.faultCode
     if msg is None:
@@ -327,11 +339,14 @@ def offline_reply(req, msg=None):
     else:
         faultString = msg
     response = dumps(Fault(faultCode, faultString))
-    req.content_type = "text/xml"
-    req.set_content_length(len(response))
-    req.write(response)
+    headers = [
+        ('Content-Length', str(len(response))),
+        ('Content-Type', "text/xml"),
+    ]
+    start_response('200 OK', headers)
+    return [response]
 
-def load_config(req):
+def load_config(environ):
     """Load configuration options
 
     Options are read from a config file. The config file location is
@@ -344,18 +359,21 @@ def load_config(req):
         - all PythonOptions (except ConfigFile) are now deprecated and support for them
           will disappear in a future version of Koji
     """
+    logger = logging.getLogger("koji")
     #get our config file
-    modpy_opts = req.get_options()
-    #cf = modpy_opts.get('ConfigFile', '/etc/koji-hub/hub.conf')
-    cf = modpy_opts.get('ConfigFile', None)
+    if 'modpy.opts' in environ:
+        modpy_opts = environ.get('modpy.opts')
+        cf = modpy_opts.get('ConfigFile', None)
+    else:
+        cf = environ.get('koji.hub.ConfigFile', '/etc/koji-hub/hub.conf')
+        modpy_opts = {}
     if cf:
         # to aid in the transition from PythonOptions to hub.conf, we only load
         # the configfile if it is explicitly configured
-        config = ConfigParser()
+        config = RawConfigParser()
         config.read(cf)
     else:
-        sys.stderr.write('Warning: configuring Koji via PythonOptions is deprecated. Use hub.conf\n')
-        sys.stderr.flush()
+        logger.warn('Warning: configuring Koji via PythonOptions is deprecated. Use hub.conf')
     cfgmap = [
         #option, type, default
         ['DBName', 'string', None],
@@ -388,10 +406,26 @@ def load_config(req):
         ['EnableFunctionDebug', 'boolean', False],
 
         ['LogLevel', 'string', 'WARNING'],
-        ['LogFormat', 'string', '%(asctime)s [%(levelname)s] m=%(method)s u=%(user_name)s %(name)s: %(message)s'],
+        ['LogFormat', 'string', '%(asctime)s [%(levelname)s] m=%(method)s u=%(user_name)s p=%(process)s r=%(remoteaddr)s %(name)s: %(message)s'],
 
         ['MissingPolicyOk', 'boolean', True],
         ['EnableMaven', 'boolean', False],
+        ['EnableWin', 'boolean', False],
+        ['EnableImageMigration', 'boolean', False],
+
+        ['RLIMIT_AS', 'string', None],
+        ['RLIMIT_CORE', 'string', None],
+        ['RLIMIT_CPU', 'string', None],
+        ['RLIMIT_DATA', 'string', None],
+        ['RLIMIT_FSIZE', 'string', None],
+        ['RLIMIT_MEMLOCK', 'string', None],
+        ['RLIMIT_NOFILE', 'string', None],
+        ['RLIMIT_NPROC', 'string', None],
+        ['RLIMIT_OFILE', 'string', None],
+        ['RLIMIT_RSS', 'string', None],
+        ['RLIMIT_STACK', 'string', None],
+
+        ['MemoryWarnThreshold', 'integer', 5000],
 
         ['LockOut', 'boolean', False],
         ['ServerOffline', 'boolean', False],
@@ -472,7 +506,11 @@ _default_policies = {
             has req_channel :: req
             is_child_task :: parent
             all :: use default
-            '''
+            ''',
+    'vm' : '''
+            has_perm admin win-admin :: allow
+            all :: deny
+           '''
 }
 
 def get_policy(opts, plugins):
@@ -508,6 +546,12 @@ class HubFormatter(logging.Formatter):
 
     def format(self, record):
         record.method = getattr(context, 'method', None)
+        if hasattr(context, 'environ'):
+            record.remoteaddr = "%s:%s" % (
+                context.environ.get('REMOTE_ADDR', '?'),
+                context.environ.get('REMOTE_PORT', '?'))
+        else:
+            record.remoteaddr = "?:?"
         if hasattr(context, 'session'):
             record.user_id = context.session.user_id
             record.session_id = context.session.id
@@ -520,8 +564,21 @@ class HubFormatter(logging.Formatter):
             record.user_name = None
         return logging.Formatter.format(self, record)
 
+def setup_logging1():
+    """Set up basic logging, before options are loaded"""
+    global log_handler
+    logger = logging.getLogger("koji")
+    logger.setLevel(logging.WARNING)
+    #stderr logging (stderr goes to httpd logs)
+    log_handler = logging.StreamHandler()
+    log_format = '%(asctime)s [%(levelname)s] SETUP p=%(process)s %(name)s: %(message)s'
+    log_handler.setFormatter(HubFormatter(log_format))
+    log_handler.setLevel(logging.DEBUG)
+    logger.addHandler(log_handler)
 
-def setup_logging(opts):
+def setup_logging2(opts):
+    global log_handler
+    """Adjust logging based on configuration options"""
     #determine log level
     level = opts['LogLevel']
     valid_levels = ('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL')
@@ -555,75 +612,123 @@ def setup_logging(opts):
     elif default is None:
         #LogLevel did not configure a default level
         logger.setLevel(logging.WARNING)
-    #for now, just stderr logging (stderr goes to httpd logs)
-    handler = logging.StreamHandler()
-    handler.setFormatter(HubFormatter(opts['LogFormat']))
-    handler.setLevel(logging.DEBUG)
-    logger.addHandler(handler)
+    #log_handler defined in setup_logging1
+    log_handler.setFormatter(HubFormatter(opts['LogFormat']))
+
+
+def load_scripts(environ):
+    """Update path and import our scripts files"""
+    global kojihub
+    scriptsdir = os.path.dirname(environ['SCRIPT_FILENAME'])
+    sys.path.insert(0, scriptsdir)
+    import kojihub
 
 
 #
 # mod_python handler
 #
 
-firstcall = True
-ready = False
-opts = {}
+def handler(req):
+    wrapper = WSGIWrapper(req)
+    return wrapper.run(application)
 
-def handler(req, profiling=False):
-    global firstcall, ready, registry, opts, plugins, policy
-    if firstcall:
-        firstcall = False
-        opts = load_config(req)
-        setup_logging(opts)
+
+def get_memory_usage():
+    pagesize = resource.getpagesize()
+    statm = [pagesize*int(y)/1024 for y in "".join(open("/proc/self/statm").readlines()).strip().split()]
+    size, res, shr, text, lib, data, dirty = statm
+    return res - shr
+
+def server_setup(environ):
+    global opts, plugins, registry, policy
+    logger = logging.getLogger('koji')
+    try:
+        setup_logging1()
+        opts = load_config(environ)
+        setup_logging2(opts)
+        load_scripts(environ)
+        koji.util.setup_rlimits(opts)
         plugins = load_plugins(opts)
         registry = get_registry(opts, plugins)
         policy = get_policy(opts, plugins)
-        ready = True
-    if not ready:
-        #this will happen on subsequent passes if an error happens in the firstcall code
-        opts['ServerOffline'] = True
-        opts['OfflineMessage'] = 'server startup error'
-    if profiling:
-        import profile, pstats, StringIO, tempfile
-        global _profiling_req
-        _profiling_req = req
-        temp = tempfile.NamedTemporaryFile()
-        profile.run("import kojixmlrpc; kojixmlrpc.handler(kojixmlrpc._profiling_req, False)", temp.name)
-        stats = pstats.Stats(temp.name)
-        strstream = StringIO.StringIO()
-        sys.stdout = strstream
-        stats.sort_stats("time")
-        stats.print_stats()
-        req.write("<pre>" + strstream.getvalue() + "</pre>")
-        _profiling_req = None
-    else:
+        koji.db.provideDBopts(database = opts["DBName"],
+                              user = opts["DBUser"],
+                              password = opts.get("DBPass",None),
+                              host = opts.get("DBHost", None))
+    except Exception:
+        tb_str = ''.join(traceback.format_exception(*sys.exc_info()))
+        logger.error(tb_str)
+        opts = {
+            'ServerOffline': True,
+            'OfflineMessage': 'server startup error',
+            }
+
+
+#
+# wsgi handler
+#
+
+firstcall = True
+
+def application(environ, start_response):
+    global firstcall
+    if firstcall:
+        server_setup(environ)
+        firstcall = False
+    # XMLRPC uses POST only. Reject anything else
+    if environ['REQUEST_METHOD'] != 'POST':
+        headers = [
+            ('Allow', 'POST'),
+        ]
+        start_response('405 Method Not Allowed', headers)
+        response = "Method Not Allowed\nThis is an XML-RPC server. Only POST requests are accepted."
+        headers = [
+            ('Content-Length', str(len(response))),
+            ('Content-Type', "text/plain"),
+        ]
+        return [response]
+    if opts.get('ServerOffline'):
+        return offline_reply(start_response, msg=opts.get("OfflineMessage", None))
+    # XXX check request length
+    # XXX most of this should be moved elsewhere
+    if 1:
         try:
-            if opts.get('ServerOffline'):
-                offline_reply(req, msg=opts.get("OfflineMessage", None))
-                return apache.OK
+            start = time.time()
+            memory_usage_at_start = get_memory_usage()
+
             context._threadclear()
             context.commit_pending = False
             context.opts = opts
             context.handlers = HandlerAccess(registry)
-            context.req = req
+            context.environ = environ
             context.policy = policy
-            koji.db.provideDBopts(database = opts["DBName"],
-                                  user = opts["DBUser"],
-                                  password = opts.get("DBPass",None),
-                                  host = opts.get("DBHost", None))
             try:
                 context.cnx = koji.db.connect()
             except Exception:
-                offline_reply(req, msg="database outage")
-                return apache.OK
+                return offline_reply(start_response, msg="database outage")
             h = ModXMLRPCRequestHandler(registry)
-            h.handle_request(req)
+            if environ['CONTENT_TYPE'] == 'application/octet-stream':
+                response = h._wrap_handler(h.handle_upload, environ)
+            else:
+                response = h._wrap_handler(h.handle_rpc, environ)
+            headers = [
+                ('Content-Length', str(len(response))),
+                ('Content-Type', "text/xml"),
+            ]
+            start_response('200 OK', headers)
             if h.traceback:
                 #rollback
                 context.cnx.rollback()
             elif context.commit_pending:
                 context.cnx.commit()
+            memory_usage_at_end = get_memory_usage()
+            if memory_usage_at_end - memory_usage_at_start > opts['MemoryWarnThreshold']:
+                paramstr = repr(getattr(context, 'params', 'UNKNOWN'))
+                if len(paramstr) > 120:
+                    paramstr = paramstr[:117] + "..."
+                h.logger.warning("Memory usage of process %d grew from %d KiB to %d KiB (+%d KiB) processing request %s with args %s" % (os.getpid(), memory_usage_at_start, memory_usage_at_end, memory_usage_at_end - memory_usage_at_start, context.method, paramstr))
+            h.logger.debug("Returning %d bytes after %f seconds", len(response),
+                        time.time() - start)
         finally:
             #make sure context gets cleaned up
             if hasattr(context,'cnx'):
@@ -632,13 +737,14 @@ def handler(req, profiling=False):
                 except Exception:
                     pass
             context._threadclear()
-    return apache.OK
+        return [response] #XXX
+
 
 def get_registry(opts, plugins):
     # Create and populate handler registry
     registry = HandlerRegistry()
-    functions = RootExports()
-    hostFunctions = HostExports()
+    functions = kojihub.RootExports()
+    hostFunctions = kojihub.HostExports()
     registry.register_instance(functions)
     registry.register_module(hostFunctions,"host")
     registry.register_function(koji.auth.login)
@@ -652,4 +758,3 @@ def get_registry(opts, plugins):
     for name in opts.get('Plugins', '').split():
         registry.register_plugin(plugins.get(name))
     return registry
-
