@@ -1,7 +1,7 @@
 # Python module
 # Common functions
 
-# Copyright (c) 2005-2007 Red Hat
+# Copyright (c) 2005-2012 Red Hat
 #
 #    Koji is free software; you can redistribute it and/or
 #    modify it under the terms of the GNU Lesser General Public
@@ -28,7 +28,9 @@ except ImportError:
     sys.stderr.flush()
 import base64
 import datetime
+import errno
 from fnmatch import fnmatch
+import httplib
 import logging
 import logging.handlers
 from koji.util import md5_constructor
@@ -38,8 +40,10 @@ import pwd
 import random
 import re
 import rpm
+import shutil
 import signal
 import socket
+import ssl.SSLCommon
 import struct
 import tempfile
 import time
@@ -49,11 +53,10 @@ import urllib2
 import urlparse
 import util
 import xmlrpclib
-from xmlrpclib import loads, Fault
 import xml.sax
 import xml.sax.handler
-import ssl.XMLRPCServerProxy
-import OpenSSL.SSL
+from xmlrpclib import loads, dumps, Fault
+#import OpenSSL.SSL
 import zipfile
 
 def _(args):
@@ -206,11 +209,25 @@ BR_STATES = Enum((
     'EXPIRED',
 ))
 
+TAG_UPDATE_TYPES = Enum((
+    'VOLUME_CHANGE',
+    'IMPORT',
+    'MANUAL',
+))
+
+CHECKSUM_TYPES = Enum((
+    'md5',
+    'sha1',
+    'sha256',
+))
+
 #PARAMETERS
 BASEDIR = '/mnt/koji'
 # default task priority
 PRIO_DEFAULT = 20
 
+## BEGIN kojikamid dup
+
 #Exceptions
 class GenericError(Exception):
     """Base class for our custom exceptions"""
@@ -224,8 +241,9 @@ class GenericError(Exception):
                 return str(self.args[0])
             except:
                 return str(self.__dict__)
+## END kojikamid dup
 
-class LockConflictError(GenericError):
+class LockError(GenericError):
     """Raised when there is a lock conflict"""
     faultCode = 1001
 
@@ -241,9 +259,12 @@ class ActionNotAllowed(GenericError):
     """Raised when the session does not have permission to take some action"""
     faultCode = 1004
 
+## BEGIN kojikamid dup
+
 class BuildError(GenericError):
     """Raised when a build fails"""
     faultCode = 1005
+## END kojikamid dup
 
 class AuthLockError(AuthError):
     """Raised when a lock prevents authentication"""
@@ -297,6 +318,11 @@ class ApplianceError(GenericError):
     """Raised when Appliance Image creation fails"""
     faultCode = 1018
 
+class ParameterError(GenericError):
+    """Raised when an rpc call receives incorrect arguments"""
+    faultCode = 1019
+
+
 class MultiCallInProgress(object):
     """
     Placeholder class to be returned by method calls when in the process of
@@ -304,6 +330,7 @@ class MultiCallInProgress(object):
     """
     pass
 
+
 #A function to get create an exception from a fault
 def convertFault(fault):
     """Convert a fault to the corresponding Exception type, if possible"""
@@ -378,12 +405,15 @@ def decode_args2(args, names, strict=True):
     ret.update(opts)
     return ret
 
+## BEGIN kojikamid dup
+
 def encode_int(n):
     """If n is too large for a 32bit signed, convert it to a string"""
     if n <= 2147483647:
         return n
     #else
     return str(n)
+## END kojikamid dup
 
 def decode_int(n):
     """If n is not an integer, attempt to convert it"""
@@ -401,17 +431,32 @@ def safe_xmlrpc_loads(s):
     except Fault, f:
         return f
 
+## BEGIN kojikamid dup
+
 def ensuredir(directory):
     """Create directory, if necessary."""
-    if os.path.isdir(directory):
-        return
-    try:
-        os.makedirs(directory)
-    except OSError:
-        #thrown when dir already exists (could happen in a race)
+    if os.path.exists(directory):
         if not os.path.isdir(directory):
-            #something else must have gone wrong
-            raise
+            raise OSError, "Not a directory: %s" % directory
+    else:
+        head, tail = os.path.split(directory)
+        if not tail and head == directory:
+            # can only happen if directory == '/' or equivalent
+            # (which obviously should not happen)
+            raise OSError, "root directory missing? %s" % directory
+        if head:
+            ensuredir(head)
+        # note: if head is blank, then we've reached the top of a relative path
+        try:
+            os.mkdir(directory)
+        except OSError:
+            #thrown when dir already exists (could happen in a race)
+            if not os.path.isdir(directory):
+                #something else must have gone wrong
+                raise
+    return directory
+
+## END kojikamid dup
 
 def daemonize():
     """Detach and run in background"""
@@ -856,7 +901,7 @@ def parse_NVRA(nvra):
 
 def is_debuginfo(name):
     """Determines if an rpm is a debuginfo rpm, based on name"""
-    if name.endswith('-debuginfo') or name.endswith('-debuginfo-common'):
+    if name.endswith('-debuginfo') or name.find('-debuginfo-') != -1:
         return True
     return False
 
@@ -868,6 +913,8 @@ def canonArch(arch):
         return 'i386'
     elif arch == 'ia32e':
         return 'x86_64'
+    elif fnmatch(arch,'ppc64le'):
+        return 'ppc64le'
     elif fnmatch(arch,'ppc64*'):
         return 'ppc64'
     elif fnmatch(arch,'sparc64*'):
@@ -876,6 +923,10 @@ def canonArch(arch):
         return 'sparc'
     elif fnmatch(arch, 'alpha*'):
         return 'alpha'
+    elif fnmatch(arch,'arm*h*'):
+        return 'armhfp'
+    elif fnmatch(arch,'arm*'):
+        return 'arm'
     else:
         return arch
 
@@ -1193,20 +1244,27 @@ def genMockConfig(name, arch, managed=False, repoid=None, tag_name=None, **opts)
     The generated config is compatible with mock >= 0.8.7
     """
     mockdir = opts.get('mockdir', '/var/lib/mock')
-    url = opts.get('url')
-    if not url:
+    if 'url' in opts:
+        from warnings import warn
+        warn('The url option for genMockConfig is deprecated', DeprecationWarning)
+        urls = [opts['url']]
+    else:
         if not (repoid and tag_name):
-            raise GenericError, "please provide a url or repo/tag"
-        topurl = opts.get('topurl')
-        if topurl:
+            raise GenericError, "please provide a repo and tag"
+        topurls = opts.get('topurls')
+        if not topurls:
+            #cli command still passes plain topurl
+            topurl = opts.get('topurl')
+            if topurl:
+                topurls = [topurl]
+        if topurls:
             #XXX - PathInfo isn't quite right for this, but it will do for now
-            pathinfo = PathInfo(topdir=topurl)
-            repodir = pathinfo.repo(repoid,tag_name)
-            url = "%s/%s" % (repodir,arch)
+            pathinfos = [PathInfo(topdir=_u) for _u in topurls]
+            urls = ["%s/%s" % (_p.repo(repoid,tag_name), arch) for _p in pathinfos]
         else:
             pathinfo = PathInfo(topdir=opts.get('topdir', '/mnt/koji'))
             repodir = pathinfo.repo(repoid,tag_name)
-            url = "file://%s/%s" % (repodir,arch)
+            urls = ["file://%s/%s" % (repodir,arch)]
     if managed:
         buildroot_id = opts.get('buildroot_id')
 
@@ -1215,14 +1273,14 @@ def genMockConfig(name, arch, managed=False, repoid=None, tag_name=None, **opts)
     config_opts = {
         'root' : name,
         'basedir' : mockdir,
-        'target_arch' : arch,
+        'target_arch' : opts.get('target_arch', arch),
         'chroothome': '/builddir',
         # Use the group data rather than a generated rpm
         'chroot_setup_cmd': 'groupinstall %s' % opts.get('install_group', 'build'),
         # don't encourage network access from the chroot
         'use_host_resolv': opts.get('use_host_resolv', False),
         # Don't let a build last more than 24 hours
-        'rpmbuild_timeout': 86400
+        'rpmbuild_timeout': opts.get('rpmbuild_timeout', 86400)
     }
 
     # bind_opts are used to mount parts (or all of) /dev if needed.
@@ -1236,10 +1294,22 @@ def genMockConfig(name, arch, managed=False, repoid=None, tag_name=None, **opts)
         etc_hosts = file('/etc/hosts')
         files['etc/hosts'] = etc_hosts.read()
         etc_hosts.close()
-    if opts.get('maven_opts', False):
-        files['etc/mavenrc'] = 'MAVEN_OPTS="%s"\n' % opts['maven_opts']
+    mavenrc = ''
+    if opts.get('maven_opts'):
+        mavenrc = 'export MAVEN_OPTS="%s"\n' % ' '.join(opts['maven_opts'])
+    if opts.get('maven_envs'):
+        for name, val in opts['maven_envs'].iteritems():
+            mavenrc += 'export %s="%s"\n' % (name, val)
+    if mavenrc:
+        files['etc/mavenrc'] = mavenrc
 
-    config_opts['yum.conf'] = """[main]
+    #generate yum.conf
+    yc_parts = ["[main]\n"]
+    # HTTP proxy for yum
+    if opts.get('yum_proxy'):
+        yc_parts.append("proxy=%s\n" % opts['yum_proxy'])
+    # Rest of the yum options
+    yc_parts.append("""\
 cachedir=/var/cache/yum
 debuglevel=1
 logfile=/var/log/yum.log
@@ -1253,8 +1323,11 @@ assumeyes=1
 
 [build]
 name=build
-baseurl=%(url)s
-""" % locals()
+""")
+    yc_parts.append("baseurl=%s\n" % urls[0])
+    for url in urls[1:]:
+        yc_parts.append("        %s\n" % url)
+    config_opts['yum.conf'] = ''.join(yc_parts)
 
     plugin_conf = {
         'ccache_enable': False,
@@ -1266,8 +1339,8 @@ baseurl=%(url)s
     macros = {
         '%_topdir' : '%s/build' % config_opts['chroothome'],
         '%_rpmfilename' : '%%{NAME}-%%{VERSION}-%%{RELEASE}.%%{ARCH}.rpm',
-        '%_host_cpu' : arch,
-        '%_host': '%s-%s' % (arch, opts.get('mockhost', 'koji-linux-gnu')),
+        '%_host_cpu' : opts.get('target_arch', arch),
+        '%_host': '%s-%s' % (opts.get('target_arch', arch), opts.get('mockhost', 'koji-linux-gnu')),
         '%vendor' : opts.get('vendor', 'Koji'),
         '%packager' : opts.get('packager', 'Koji'),
         '%distribution': opts.get('distribution', 'Unknown')
@@ -1349,13 +1422,11 @@ def openRemoteFile(relpath, topurl=None, topdir=None):
     on options"""
     if topurl:
         url = "%s/%s" % (topurl, relpath)
-        fo = urllib2.urlopen(url)
-
-        # horrible hack because urlopen() returns a wrapped HTTPConnection
-        # object which doesn't implement fileno()
-        def urlopen_fileno(fileobj):
-            return lambda: fileobj.fp._sock.fp.fileno()
-        fo.fileno = urlopen_fileno(fo)
+        src = urllib2.urlopen(url)
+        fo = tempfile.TemporaryFile()
+        shutil.copyfileobj(src, fo)
+        src.close()
+        fo.seek(0)
     elif topdir:
         fn = "%s/%s" % (topdir, relpath)
         fo = open(fn)
@@ -1368,31 +1439,59 @@ class PathInfo(object):
     # ASCII numbers and upper- and lower-case letter for use in tmpdir()
     ASCII_CHARS = [chr(i) for i in range(48, 58) + range(65, 91) + range(97, 123)]
 
-    def __init__(self,topdir=None):
-        if topdir is None:
-            self.topdir = BASEDIR
-        else:
-            self.topdir = topdir
+    def __init__(self, topdir=None):
+        self._topdir = topdir
+
+    def topdir(self):
+        if self._topdir is None:
+            self._topdir = str(BASEDIR)
+        return self._topdir
+
+    def _set_topdir(self, topdir):
+        self._topdir = topdir
+
+    topdir = property(topdir, _set_topdir)
+
+    def volumedir(self, volume):
+        if volume == 'DEFAULT' or volume is None:
+            return self.topdir
+        #else
+        return self.topdir + ("/vol/%s" % volume)
 
     def build(self,build):
         """Return the directory where a build belongs"""
-        return self.topdir + ("/packages/%(name)s/%(version)s/%(release)s" % build)
+        return self.volumedir(build.get('volume_name')) + ("/packages/%(name)s/%(version)s/%(release)s" % build)
 
-    def mavenbuild(self, build, maveninfo):
-        """Return the directory where the Maven build exists in the global store (/mnt/koji/maven2)"""
+    def mavenbuild(self, build):
+        """Return the directory where the Maven build exists in the global store (/mnt/koji/packages)"""
+        return self.build(build) + '/maven'
+
+    def mavenrepo(self, maveninfo):
+        """Return the relative path to the artifact directory in the repo"""
         group_path = maveninfo['group_id'].replace('.', '/')
         artifact_id = maveninfo['artifact_id']
         version = maveninfo['version']
-        release = build['release']
-        return self.topdir + ("/maven2/%(group_path)s/%(artifact_id)s/%(version)s/%(release)s" % locals())
+        return "%(group_path)s/%(artifact_id)s/%(version)s" % locals()
 
-    def mavenrepo(self, build, maveninfo):
-        """Return the directory where the Maven artifact exists in the per-tag Maven repo
-        (/mnt/koji/repos/tag-name/repo-id/maven2/)"""
-        group_path = maveninfo['group_id'].replace('.', '/')
-        artifact_id = maveninfo['artifact_id']
-        version = maveninfo['version']
-        return self.topdir + ("/maven2/%(group_path)s/%(artifact_id)s/%(version)s" % locals())
+    def mavenfile(self, maveninfo):
+        """Return the relative path to the artifact in the repo"""
+        return self.mavenrepo(maveninfo) + '/' + maveninfo['filename']
+
+    def winbuild(self, build):
+        """Return the directory where the Windows build exists"""
+        return self.build(build) + '/win'
+
+    def winfile(self, wininfo):
+        """Return the relative path from the winbuild directory where the
+           file identified by wininfo is located."""
+        filepath = wininfo['filename']
+        if wininfo['relpath']:
+            filepath = wininfo['relpath'] + '/' + filepath
+        return filepath
+
+    def imagebuild(self, build):
+        """Return the directory where the image for the build are stored"""
+        return self.build(build) + '/images'
 
     def rpm(self,rpminfo):
         """Return the path (relative to build_dir) where an rpm belongs"""
@@ -1422,18 +1521,6 @@ class PathInfo(object):
         """Return the relative path for the task work directory"""
         return "tasks/%s/%s" % (task_id % 10000, task_id)
 
-    def livecdRelPath(self, image_id):
-        """Return the relative path for the livecd image directory"""
-        return os.path.join('livecd', str(image_id % 10000), str(image_id))
-
-    def applianceRelPath(self, image_id):
-        """Return the relative path for the appliance image directory"""
-        return os.path.join('appliance', str(image_id % 10000), str(image_id))
-
-    def imageFinalPath(self):
-        """Return the absolute path to where completed images can be found"""
-        return os.path.join(self.topdir, 'images')
-
     def work(self):
         """Return the work dir"""
         return self.topdir + '/work'
@@ -1476,24 +1563,51 @@ class ClientSession(object):
             opts = {}
         else:
             opts = opts.copy()
-        self.opts = opts
-        self.proxyOpts = {'allow_none':1}
-        if self.opts.get('debug_xmlrpc'):
-            self.proxyOpts['verbose'] = 1
-        if self.opts.get('certs'):
-            self.proxyOpts['certs'] = self.opts['certs']
-            self.proxyClass = ssl.XMLRPCServerProxy.PlgXMLRPCServerProxy
-        else:
-            self.proxyClass = xmlrpclib.ServerProxy
-        if self.opts.get('timeout'):
-            self.proxyOpts['timeout'] = self.opts['timeout']
         self.baseurl = baseurl
-        self.origurl = None
+        self.opts = opts
+        self._connection = None
+        self._setup_connection()
         self.setSession(sinfo)
         self.multicall = False
         self._calls = []
         self.logger = logging.getLogger('koji')
 
+    def _setup_connection(self):
+        uri = urlparse.urlsplit(self.baseurl)
+        scheme = uri[0]
+        self._host, _port = urllib.splitport(uri[1])
+        self.explicit_port = bool(_port)
+        self._path = uri[2]
+        default_port = 80
+        if self.opts.get('certs'):
+            ctx = ssl.SSLCommon.CreateSSLContext(self.opts['certs'])
+            cnxOpts = {'ssl_context' : ctx}
+            cnxClass = ssl.SSLCommon.PlgHTTPSConnection
+            default_port = 443
+        elif scheme == 'https':
+            cnxOpts = {}
+            cnxClass = httplib.HTTPSConnection
+            default_port = 443
+        elif scheme == 'http':
+            cnxOpts = {}
+            cnxClass = httplib.HTTPConnection
+        else:
+            raise IOError, "unsupported XML-RPC protocol"
+        # set a default 12 hour connection timeout.
+        # Some Koji operations can take a long time to return, but after 12
+        # hours we can assume something is seriously wrong.
+        timeout = self.opts.setdefault('timeout',  60 * 60 * 12)
+        self._timeout_compat = False
+        if timeout:
+            if sys.version_info[:3] < (2, 6, 0) and 'ssl_context' not in cnxOpts:
+                self._timeout_compat = True
+            else:
+                cnxOpts['timeout'] = timeout
+        self._port = (_port and int(_port) or default_port)
+        self._cnxOpts = cnxOpts
+        self._cnxClass = cnxClass
+        self._close_connection()
+
     def setSession(self,sinfo):
         """Set the session info
 
@@ -1501,22 +1615,12 @@ class ClientSession(object):
         if sinfo is None:
             self.logged_in = False
             self.callnum = None
-            # undo state changes made by ssl_login()
-            if self.origurl:
-                self.baseurl = self.origurl
-                self.origurl = None
-            self.opts.pop('certs', None)
-            self.proxyOpts.pop('certs', None)
-            self.opts.pop('timeout', None)
-            self.proxyOpts.pop('timeout', None)
-            self.proxyClass = xmlrpclib.ServerProxy
-            url = self.baseurl
+            # do we need to do anything else here?
+            self._setup_connection()
         else:
             self.logged_in = True
             self.callnum = 0
-            url = "%s?%s" %(self.baseurl,urllib.urlencode(sinfo))
         self.sinfo = sinfo
-        self.proxy = self.proxyClass(url,**self.proxyOpts)
 
     def login(self,opts=None):
         sinfo = self.callMethod('login',self.opts['user'], self.opts['password'],opts)
@@ -1558,7 +1662,7 @@ class ClientSession(object):
             # We're trying to log ourself in.  Connect using existing credentials.
             cprinc = ccache.principal()
 
-        sprinc = krbV.Principal(name=self._serverPrincipal(), context=ctx)
+        sprinc = krbV.Principal(name=self._serverPrincipal(cprinc), context=ctx)
 
         ac = krbV.AuthContext(context=ctx)
         ac.flags = krbV.KRB5_AUTH_CONTEXT_DO_SEQUENCE|krbV.KRB5_AUTH_CONTEXT_DO_TIME
@@ -1595,54 +1699,55 @@ class ClientSession(object):
 
         return True
 
-    def _serverPrincipal(self):
+    def _serverPrincipal(self, cprinc):
         """Get the Kerberos principal of the server we're connecting
-        to, based on baseurl.  Assume the last two components of the
-        server name are the Kerberos realm."""
-        servername = urlparse.urlparse(self.baseurl)[1]
-        portspec = servername.find(':')
-        if portspec != -1:
-            servername = servername[:portspec]
+        to, based on baseurl."""
+        servername = self._host
+        #portspec = servername.find(':')
+        #if portspec != -1:
+        #    servername = servername[:portspec]
+        realm = cprinc.realm
+        service = self.opts.get('krbservice', 'host')
 
-        parts = servername.split('.')
-        if len(parts) < 2:
-            domain = servername.upper()
-        else:
-            domain = '.'.join(parts[-2:]).upper()
-
-        return 'host/%s@%s' % (servername, domain)
+        return '%s/%s@%s' % (service, servername, realm)
 
     def ssl_login(self, cert, ca, serverca, proxyuser=None):
-        if not self.baseurl.startswith('https:'):
-            self.origurl = self.baseurl
-            self.baseurl = self.baseurl.replace('http:', 'https:', 1)
-        
         certs = {}
         certs['key_and_cert'] = cert
         certs['ca_cert'] = ca
         certs['peer_ca_cert'] = serverca
 
+        ctx = ssl.SSLCommon.CreateSSLContext(certs)
+        self._cnxOpts = {'ssl_context' : ctx}
         # 60 second timeout during login
-        # Append /login to the URL so we can only require client certs to be sent on login requests
-        self.proxy = ssl.XMLRPCServerProxy.PlgXMLRPCServerProxy(self.baseurl + '/ssllogin', certs, timeout=60, **self.proxyOpts)
-        sinfo = self.callMethod('sslLogin', proxyuser)
+        old_timeout = self._cnxOpts.get('timeout')
+        self._cnxOpts['timeout'] = 60
+        try:
+            self._cnxClass = ssl.SSLCommon.PlgHTTPSConnection
+            if self._port == 80 and not self.explicit_port:
+                self._port = 443
+            sinfo = self.callMethod('sslLogin', proxyuser)
+        finally:
+            if old_timeout is None:
+                del self._cnxOpts['timeout']
+            else:
+                self._cnxOpts['timeout'] = old_timeout
         if not sinfo:
             raise AuthError, 'unable to obtain a session'
 
-        self.proxyClass = ssl.XMLRPCServerProxy.PlgXMLRPCServerProxy
-        self.opts['certs'] = self.proxyOpts['certs'] = certs
-        # 12 hour connection timeout.  Some Koji operations can take a long time to return,
-        # but after 12 hours we can assume something is seriously wrong.
-        self.opts['timeout'] = self.proxyOpts['timeout'] = 60 * 60 * 12
+        self.opts['certs'] = certs
         self.setSession(sinfo)
 
         return True
-        
+
     def logout(self):
         if not self.logged_in:
             return
         try:
-            self.proxy.logout()
+            # bypass _callMethod (no retries)
+            # XXX - is that really what we want?
+            handler, headers, request = self._prepCall('logout', ())
+            self._sendCall(handler, headers, request)
         except AuthExpired:
             #this can happen when an exclusive session is forced
             pass
@@ -1672,30 +1777,125 @@ class ClientSession(object):
         """compatibility wrapper for _callMethod"""
         return self._callMethod(name, args, opts)
 
-    def _callMethod(self, name, args, kwargs):
+    def _prepCall(self, name, args, kwargs=None):
         #pass named opts in a way the server can understand
+        if kwargs is None:
+            kwargs = {}
+        if name == 'rawUpload':
+            return self._prepUpload(*args, **kwargs)
         args = encode_args(*args,**kwargs)
+        if self.logged_in:
+            sinfo = self.sinfo.copy()
+            sinfo['callnum'] = self.callnum
+            self.callnum += 1
+            handler = "%s?%s" % (self._path, urllib.urlencode(sinfo))
+        elif name == 'sslLogin':
+            handler = self._path + '/ssllogin'
+        else:
+            handler = self._path
+        request = dumps(args, name, allow_none=1)
+        headers = [
+            # connection class handles Host
+            ('User-Agent', 'koji/1.7'),  #XXX
+            ('Content-Type', 'text/xml'),
+            ('Content-Length', len(request)),
+        ]
+        return handler, headers, request
+
+    def _sendCall(self, handler, headers, request):
+        # handle expired connections
+        for i in (0, 1):
+            try:
+                return self._sendOneCall(handler, headers, request)
+            except socket.error, e:
+                self._close_connection()
+                if i or getattr(e, 'errno', None) not in (errno.ECONNRESET, errno.ECONNABORTED, errno.EPIPE):
+                    raise
+            except httplib.BadStatusLine:
+                self._close_connection()
+                if i:
+                    raise
+
+
+    def _sendOneCall(self, handler, headers, request):
+        cnx = self._get_connection()
+        if self.opts.get('debug_xmlrpc', False):
+            cnx.set_debuglevel(1)
+        cnx.putrequest('POST', handler)
+        for n, v in headers:
+            cnx.putheader(n, v)
+        cnx.endheaders()
+        cnx.send(request)
+        response = cnx.getresponse()
+        try:
+            ret = self._read_xmlrpc_response(response, handler)
+        finally:
+            response.close()
+        return ret
+
+    def _get_connection(self):
+        key = (self._cnxClass, self._host, self._port)
+        if self._connection and self.opts.get('keepalive'):
+            if key == self._connection[0]:
+                cnx = self._connection[1]
+                if getattr(cnx, 'sock', None):
+                    return cnx
+        cnx = self._cnxClass(self._host, self._port, **self._cnxOpts)
+        self._connection = (key, cnx)
+        if self._timeout_compat:
+            # in python < 2.6 httplib does not support the timeout option
+            # but socket supports it since 2.3
+            cnx.connect()
+            cnx.sock.settimeout(self.opts['timeout'])
+        return cnx
+
+    def _close_connection(self):
+        if self._connection:
+            self._connection[1].close()
+            self._connection = None
+
+    def _read_xmlrpc_response(self, response, handler=''):
+        #XXX honor debug_xmlrpc
+        if response.status != 200:
+            if (response.getheader("content-length", 0)):
+                response.read()
+            raise xmlrpclib.ProtocolError(self._host + handler,
+                        response.status, response.reason, response.msg)
+        p, u = xmlrpclib.getparser()
+        while True:
+            chunk = response.read(8192)
+            if not chunk:
+                break
+            if self.opts.get('debug_xmlrpc', False):
+                print "body: %r" % chunk
+            p.feed(chunk)
+        p.close()
+        result = u.close()
+        if len(result) == 1:
+            result = result[0]
+        return result
+
+    def _callMethod(self, name, args, kwargs=None):
+        """Make a call to the hub with retries and other niceties"""
 
         if self.multicall:
+            if kwargs is None:
+                kwargs = {}
+            args = encode_args(*args, **kwargs)
             self._calls.append({'methodName': name, 'params': args})
             return MultiCallInProgress
         else:
-            if self.logged_in:
-                sinfo = self.sinfo.copy()
-                sinfo['callnum'] = self.callnum
-                self.callnum += 1
-                url = "%s?%s" %(self.baseurl,urllib.urlencode(sinfo))
-                proxy = self.proxyClass(url,**self.proxyOpts)
-            else:
-                proxy = self.proxy
+            handler, headers, request = self._prepCall(name, args, kwargs)
             tries = 0
+            self.retries = 0
             debug = self.opts.get('debug',False)
             max_retries = self.opts.get('max_retries',30)
             interval = self.opts.get('retry_interval',20)
             while True:
                 tries += 1
+                self.retries += 1
                 try:
-                    return proxy.__getattr__(name)(*args)
+                    return self._sendCall(handler, headers, request)
                 #basically, we want to retry on most errors, with a few exceptions
                 #  - faults (this means the call completed and failed)
                 #  - SystemExit, KeyboardInterrupt
@@ -1708,8 +1908,7 @@ class ClientSession(object):
                     if isinstance(err, ServerOffline):
                         if self.opts.get('offline_retry',False):
                             secs = self.opts.get('offline_retry_interval', interval)
-                            if debug:
-                                self.logger.debug("Server offline. Retrying in %i seconds" % secs)
+                            self.logger.debug("Server offline. Retrying in %i seconds", secs)
                             time.sleep(secs)
                             #reset try count - this isn't a typical error, this is a running server
                             #correctly reporting an outage
@@ -1720,6 +1919,7 @@ class ClientSession(object):
                     #(depending on the python version, these may or may not be subclasses of Exception)
                     raise
                 except Exception, e:
+                    self._close_connection()
                     if not self.logged_in:
                         #in the past, non-logged-in sessions did not retry. For compatibility purposes
                         #this behavior is governed by the anon_retry opt.
@@ -1728,8 +1928,10 @@ class ClientSession(object):
                     if tries > max_retries:
                         raise
                     #otherwise keep retrying
-                    if debug:
-                        self.logger.debug("Try #%d for call %d (%s) failed: %s" % (tries, self.callnum, name, e))
+                    if self.logger.isEnabledFor(logging.DEBUG):
+                        tb_str = ''.join(traceback.format_exception(*sys.exc_info()))
+                        self.logger.debug(tb_str)
+                    self.logger.info("Try #%d for call %d (%s) failed: %s", tries, self.callnum, name, e)
                 time.sleep(interval)
             #not reached
 
@@ -1769,13 +1971,100 @@ class ClientSession(object):
         #    raise AttributeError, "no attribute %r" % name
         return VirtualMethod(self._callMethod,name)
 
-    def uploadWrapper(self, localfile, path, name=None, callback=None, blocksize=1048576):
-        """upload a file in chunks using the uploadFile call"""
-        # XXX - stick in a config or something
-        start=time.time()
-        retries=3
+    def fastUpload(self, localfile, path, name=None, callback=None, blocksize=1048576):
+        if not self.logged_in:
+            raise ActionNotAllowed, 'You must be logged in to upload files'
         if name is None:
             name = os.path.basename(localfile)
+        self.logger.debug("Fast upload: %s to %s/%s", localfile, path, name)
+        size = os.stat(localfile).st_size
+        fo = file(localfile, 'rb')
+        ofs = 0
+        size = os.path.getsize(localfile)
+        start = time.time()
+        if callback:
+            callback(0, size, 0, 0, 0)
+        problems = False
+        full_chksum = util.adler32_constructor()
+        while True:
+            lap = time.time()
+            chunk = fo.read(blocksize)
+            if not chunk:
+                break
+            result = self._callMethod('rawUpload', (chunk, ofs, path, name))
+            if self.retries > 1:
+                problems = True
+            hexdigest = util.adler32_constructor(chunk).hexdigest()
+            full_chksum.update(chunk)
+            if result['size'] != len(chunk):
+                raise GenericError, "server returned wrong chunk size: %s != %s" % (result['size'], len(chunk))
+            if result['hexdigest'] != hexdigest:
+                raise GenericError, 'upload checksum failed: %s != %s' \
+                        % (result['hexdigest'], hexdigest)
+            ofs += len(chunk)
+            now = time.time()
+            t1 = max(now - lap, 0.00001)
+            t2 = max(now - start, 0.00001)
+            # max is to prevent possible divide by zero in callback function
+            if callback:
+                callback(ofs, size, len(chunk), t1, t2)
+        if ofs != size:
+            self.logger.error("Local file changed size: %s, %s -> %s", localfile, size, ofs)
+        chk_opts = {}
+        if problems:
+            chk_opts['verify'] = 'adler32'
+        result = self._callMethod('checkUpload', (path, name), chk_opts)
+        if int(result['size']) != ofs:
+            raise koji.GenericError, "Uploaded file is wrong length: %s/%s, %s != %s" \
+                    % (path, name, result['sumlength'], ofs)
+        if problems and result['hexdigest'] != full_chksum.hexdigest():
+            raise koji.GenericError, "Uploaded file has wrong checksum: %s/%s, %s != %s" \
+                    % (path, name, result['hexdigest'], full_chksum.hexdigest())
+        self.logger.debug("Fast upload: %s complete. %i bytes in %.1f seconds", localfile, size, t2)
+
+    def _prepUpload(self, chunk, offset, path, name, verify="adler32", overwrite=False):
+        """prep a rawUpload call"""
+        if not self.logged_in:
+            raise ActionNotAllowed, "you must be logged in to upload"
+        args = self.sinfo.copy()
+        args['callnum'] = self.callnum
+        args['filename'] = name
+        args['filepath'] = path
+        args['fileverify'] = verify
+        args['offset'] = str(offset)
+        if overwrite:
+            args['overwrite'] = "1"
+        size = len(chunk)
+        self.callnum += 1
+        handler = "%s?%s" % (self._path, urllib.urlencode(args))
+        headers = [
+            ('User-Agent', 'koji/1.7'),  #XXX
+            ("Content-Type", "application/octet-stream"),
+            ("Content-length", str(size)),
+        ]
+        request = chunk
+        return handler, headers, request
+
+    def uploadWrapper(self, localfile, path, name=None, callback=None, blocksize=1048576):
+        """upload a file in chunks using the uploadFile call"""
+        if self.opts.get('use_fast_upload'):
+            self.fastUpload(localfile, path, name, callback, blocksize)
+            return
+        if name is None:
+            name = os.path.basename(localfile)
+
+        # check if server supports fast upload
+        try:
+            check = self._callMethod('checkUpload', (path, name))
+        except GenericError:
+            pass
+        else:
+            self.fastUpload(localfile, path, name, callback, blocksize)
+            return
+
+        start=time.time()
+        # XXX - stick in a config or something
+        retries=3
         fo = file(localfile, "r")  #specify bufsize?
         totalsize = os.path.getsize(localfile)
         ofs = 0
@@ -1930,13 +2219,17 @@ def formatTimeLong(value):
 
 def buildLabel(buildInfo, showEpoch=False):
     """Format buildInfo (dict) into a descriptive label."""
-    epoch = buildInfo['epoch']
+    epoch = buildInfo.get('epoch')
     if showEpoch and epoch != None:
         epochStr = '%i:' % epoch
     else:
         epochStr = ''
-    return '%s%s-%s-%s' % (epochStr, buildInfo['package_name'],
-                           buildInfo['version'], buildInfo['release'])
+    name = buildInfo.get('package_name')
+    if not name:
+        name = buildInfo.get('name')
+    return '%s%s-%s-%s' % (epochStr, name,
+                           buildInfo.get('version'),
+                           buildInfo.get('release'))
 
 def _module_info(url):
     module_info = ''
@@ -1957,6 +2250,12 @@ def _module_info(url):
         return '%s:%s' % (repo_info, rev_info)
 
 def taskLabel(taskInfo):
+    try:
+        return _taskLabel(taskInfo)
+    except Exception:
+        return "malformed task"
+
+def _taskLabel(taskInfo):
     """Format taskInfo (dict) into a descriptive label."""
     method = taskInfo['method']
     arch = taskInfo['arch']
@@ -1984,12 +2283,22 @@ def taskLabel(taskInfo):
             extra = build_tag['name']
     elif method == 'wrapperRPM':
         if taskInfo.has_key('request'):
-            build_tag = taskInfo['request'][1]
+            build_target = taskInfo['request'][1]
             build = taskInfo['request'][2]
             if build:
-                extra = '%s, %s' % (build_tag['name'], buildLabel(build))
+                extra = '%s, %s' % (build_target['name'], buildLabel(build))
             else:
-                extra = build_tag['name']
+                extra = build_target['name']
+    elif method == 'winbuild':
+        if taskInfo.has_key('request'):
+            vm = taskInfo['request'][0]
+            url = taskInfo['request'][1]
+            target = taskInfo['request'][2]
+            module_info = _module_info(url)
+            extra = '%s, %s' % (target, module_info)
+    elif method == 'vmExec':
+        if taskInfo.has_key('request'):
+            extra = taskInfo['request'][0]
     elif method == 'buildNotification':
         if taskInfo.has_key('request'):
             build = taskInfo['request'][1]
@@ -2022,10 +2331,31 @@ def taskLabel(taskInfo):
                 nvrs = taskInfo['request'][2]
                 if isinstance(nvrs, list):
                     extra += ', ' + ', '.join(nvrs)
-    elif method == 'createLiveCD':
+    elif method in ('livecd', 'appliance', 'image'):
         if taskInfo.has_key('request'):
-            arch, target, ksfile = taskInfo['request'][:3]
-            extra = '%s, %s, %s' % (target, arch, os.path.basename(ksfile))
+            stuff = taskInfo['request']
+            if method == 'image':
+                kickstart = os.path.basename(stuff[-1]['kickstart'])
+            else:
+                kickstart = os.path.basename(stuff[4])
+            extra = '%s, %s-%s, %s' % (stuff[3], stuff[0], stuff[1], kickstart)
+    elif method in ('createLiveCD', 'createAppliance', 'createImage'):
+        if taskInfo.has_key('request'):
+            stuff = taskInfo['request']
+            if method == 'createImage':
+                kickstart = os.path.basename(stuff[-1]['kickstart'])
+            else:
+                kickstart = os.path.basename(stuff[7])
+            extra = '%s, %s-%s-%s, %s, %s' % (stuff[4]['name'], stuff[0],
+                stuff[1], stuff[2], kickstart, stuff[3])
+    elif method == 'restart':
+        if taskInfo.has_key('request'):
+            host = taskInfo['request'][0]
+            extra = host['name']
+    elif method == 'restartVerify':
+        if taskInfo.has_key('request'):
+            task_id, host = taskInfo['request'][:2]
+            extra = host['name']
 
     if extra:
         return '%s (%s)' % (method, extra)
@@ -2044,7 +2374,7 @@ def fixEncoding(value, fallback='iso8859-15'):
     encoded in the 'fallback' charset.
     """
     if not value:
-        return value
+        return ''
 
     if isinstance(value, unicode):
         # value is already unicode, so just convert it
