@@ -1,5 +1,5 @@
 # authentication module
-# Copyright (c) 2005-2007 Red Hat
+# Copyright (c) 2005-2012 Red Hat
 #
 #    Koji is free software; you can redistribute it and/or
 #    modify it under the terms of the GNU Lesser General Public
@@ -26,7 +26,6 @@ import krbV
 import koji
 import cgi      #for parse_qs
 from context import context
-from mod_python import apache
 
 # 1 - load session if provided
 #       - check uri for session id
@@ -36,6 +35,19 @@ from mod_python import apache
 #       - maybe in two steps
 #       -
 
+
+RetryWhitelist = [
+    'host.taskWait',
+    'host.taskUnwait',
+    'host.taskSetWait',
+    'host.updateHost',
+    'host.setBuildRootState',
+    'repoExpire',
+    'repoDelete',
+    'repoProblem',
+]
+
+
 class Session(object):
 
     def __init__(self,args=None,hostip=None):
@@ -52,14 +64,15 @@ class Session(object):
         self.callnum = None
         #get session data from request
         if args is None:
-            req = getattr(context,'req',None)
-            args = getattr(req,'args',None)
+            environ = getattr(context,'environ',{})
+            args = environ.get('QUERY_STRING','')
             if not args:
                 self.message = 'no session args'
                 return
             args = cgi.parse_qs(args,strict_parsing=True)
         if hostip is None:
-            hostip = context.req.get_remote_host(apache.REMOTE_NOLOOKUP)
+            hostip = context.environ['REMOTE_ADDR']
+            #XXX - REMOTE_ADDR not promised by wsgi spec
             if hostip == '127.0.0.1':
                 hostip = socket.gethostbyname(socket.gethostname())
         try:
@@ -73,8 +86,19 @@ class Session(object):
             callnum = None
         #lookup the session
         c = context.cnx.cursor()
-        fields = ('user_id','authtype','expired','start_time','update_time',
-                  'master','exclusive','callnum')
+        fields = {
+            'authtype': 'authtype',
+            'callnum': 'callnum',
+            'exclusive': 'exclusive',
+            'expired': 'expired',
+            'master': 'master',
+            'start_time': 'start_time',
+            'update_time': 'update_time',
+            'EXTRACT(EPOCH FROM start_time)': 'start_ts',
+            'EXTRACT(EPOCH FROM update_time)': 'update_ts',
+            'user_id': 'user_id',
+            }
+        fields, aliases = zip(*fields.items())
         q = """
         SELECT %s FROM sessions
         WHERE id = %%(id)i
@@ -86,7 +110,7 @@ class Session(object):
         row = c.fetchone()
         if not row:
             raise koji.AuthError, 'Invalid session or bad credentials'
-        session_data = dict(zip(fields,row))
+        session_data = dict(zip(aliases, row))
         #check for expiration
         if session_data['expired']:
             raise koji.AuthExpired, 'session "%i" has expired' % id
@@ -110,9 +134,11 @@ class Session(object):
                     #We only schedule a commit for dml operations, so if we find the
                     #callnum in the db then a previous attempt succeeded but failed to
                     #return. Data was changed, so we cannot simply try the call again.
-                    raise koji.RetryError, \
-                        "unable to retry call %d (method %s) for session %d" \
-                        % (callnum, getattr(context, 'method', 'UNKNOWN'), id)
+                    method = getattr(context, 'method', 'UNKNOWN')
+                    if method not in RetryWhitelist:
+                        raise koji.RetryError, \
+                            "unable to retry call %d (method %s) for session %d" \
+                            % (callnum, method, id)
 
         # read user data
         #historical note:
@@ -234,7 +260,8 @@ class Session(object):
             raise koji.GenericError, "Already logged in"
         hostip = opts.get('hostip')
         if hostip is None:
-            hostip = context.req.get_remote_host(apache.REMOTE_NOLOOKUP)
+            hostip = context.environ['REMOTE_ADDR']
+            #XXX - REMOTE_ADDR not promised by wsgi spec
             if hostip == '127.0.0.1':
                 hostip = socket.gethostbyname(socket.gethostname())
 
@@ -303,7 +330,8 @@ class Session(object):
 
         self.checkLoginAllowed(user_id)
 
-        hostip = context.req.connection.remote_ip
+        hostip = context.environ['REMOTE_ADDR']
+        #XXX - REMOTE_ADDR not promised by wsgi spec
         if hostip == '127.0.0.1':
             hostip = socket.gethostbyname(socket.gethostname())
 
@@ -328,15 +356,13 @@ class Session(object):
         # See: http://lists.planet-lab.org/pipermail/devel-community/2005-June/001084.html
         # local_ip seems to always be set to the same value as remote_ip,
         # so get the local ip via a different method
-        # local_ip = context.req.connection.local_ip
-        local_ip = socket.gethostbyname(context.req.hostname)
-        remote_ip = context.req.connection.remote_ip
+        local_ip = socket.gethostbyname(context.environ['SERVER_NAME'])
+        remote_ip = context.environ['REMOTE_ADDR']
+        #XXX - REMOTE_ADDR not promised by wsgi spec
 
         # it appears that calling setports() with *any* value results in authentication
         # failing with "Incorrect net address", so return 0 (which prevents
         # python-krbV from calling setports())
-        # local_port = context.req.connection.local_addr[1]
-        # remote_port = context.req.connection.remote_addr[1]
         local_port = 0
         remote_port = 0
 
@@ -346,23 +372,19 @@ class Session(object):
         if self.logged_in:
             raise koji.AuthError, "Already logged in"
 
-        # populate standard CGI variables
-        context.req.add_common_vars()
-        env = context.req.subprocess_env
-    
-        if env.get('HTTPS') != 'on':
-            raise koji.AuthError, 'cannot call sslLogin() via a non-https connection'
+        if context.environ['wsgi.url_scheme'] != 'https':
+            raise koji.AuthError, 'cannot call sslLogin() via a non-https connection: %s' % context.environ['wsgi.url_scheme']
 
-        if env.get('SSL_CLIENT_VERIFY') != 'SUCCESS':
-            raise koji.AuthError, 'could not verify client: %s' % env.get('SSL_CLIENT_VERIFY')
+        if context.environ.get('SSL_CLIENT_VERIFY') != 'SUCCESS':
+            raise koji.AuthError, 'could not verify client: %s' % context.environ.get('SSL_CLIENT_VERIFY')
 
         name_dn_component = context.opts.get('DNUsernameComponent', 'CN')
-        client_name = env.get('SSL_CLIENT_S_DN_%s' % name_dn_component)
+        client_name = context.environ.get('SSL_CLIENT_S_DN_%s' % name_dn_component)
         if not client_name:
             raise koji.AuthError, 'unable to get user information (%s) from client certificate' % name_dn_component
 
         if proxyuser:
-            client_dn = env.get('SSL_CLIENT_S_DN')
+            client_dn = context.environ.get('SSL_CLIENT_S_DN')
             proxy_dns = [dn.strip() for dn in context.opts.get('ProxyDNs', '').split('|')]
             if client_dn in proxy_dns:
                 # the SSL-authenticated user authorized to login other users
@@ -386,8 +408,9 @@ class Session(object):
                 raise koji.AuthError, 'Unknown user: %s' % username
 
         self.checkLoginAllowed(user_id)
-            
-        hostip = context.req.connection.remote_ip
+
+        hostip = context.environ['REMOTE_ADDR']
+        #XXX - REMOTE_ADDR not promised by wsgi spec
         if hostip == '127.0.0.1':
             hostip = socket.gethostbyname(socket.gethostname())
 
@@ -595,6 +618,20 @@ class Session(object):
 
         return user_id
 
+    def setKrbPrincipal(self, name, krb_principal):
+        usertype = koji.USERTYPES['NORMAL']
+        status = koji.USER_STATUS['NORMAL']
+        update = """UPDATE users SET krb_principal = %(krb_principal)s WHERE name = %(name)s AND usertype = %(usertype)i AND status = %(status)i RETURNING users.id"""
+        cursor = context.cnx.cursor()
+        cursor.execute(update, locals())
+        r = cursor.fetchall()
+        if len(r) != 1:
+            context.cnx.rollback()
+            raise koji.AuthError, 'could not automatically associate Kerberos Principal with existing user %s' % (name,)
+        else:
+            context.cnx.commit()
+            return r[0][0]
+
     def createUserFromKerberos(self, krb_principal):
         """Create a new user, based on the Kerberos principal.  Their
         username will be everything before the "@" in the principal.
@@ -604,7 +641,19 @@ class Session(object):
             raise koji.AuthError, 'invalid Kerberos principal: %s' % krb_principal
         user_name = krb_principal[:atidx]
 
-        return self.createUser(user_name, krb_principal=krb_principal)
+        # check if user already exists
+        c = context.cnx.cursor()
+        q = """SELECT krb_principal FROM users
+        WHERE name = %(user_name)s"""
+        c.execute(q,locals())
+        r = c.fetchone()
+        if not r:
+            return self.createUser(user_name, krb_principal=krb_principal)
+        else:
+            existing_user_krb = r[0]
+            if existing_user_krb is not None:
+                raise koji.AuthError, 'user %s already associated with other Kerberos principal: %s' % (user_name, existing_user_krb)
+            return self.setKrbPrincipal(user_name, krb_principal)
 
 def get_user_groups(user_id):
     """Get user groups
